"""DEPRECATED.

From https://docs.kili-technology.com/reference/graphql-api#enums.
"""

from typing import Literal

LicenseType = Literal[
    "DISCOVERY",
    "FREEMIUM",
    "PAID",
]


LockType = Literal[
    "REVIEW",
    "DEFAULT",
]


NotificationStatus = Literal[
    "FAILURE",
    "PENDING",
    "SUCCESS",
]


OrganizationRole = Literal[
    "ADMIN",
    "USER",
]


ProjectRole = Literal[
    "ADMIN",
    "TEAM_MANAGER",
    "REVIEWER",
    "LABELER",
]


ProjectType = Literal[
    "IMAGE_CLASSIFICATION_SINGLE",
    "IMAGE_CLASSIFICATION_MULTI",
    "IMAGE_OBJECT_DETECTION_RECTANGLE",
    "IMAGE_OBJECT_DETECTION_POLYGON",
    "IMAGE_OBJECT_DETECTION_SEMANTIC",
    "IMAGE_POSE_ESTIMATION",
    "OCR",
    "PDF_CLASSIFICATION_SINGLE",
    "PDF_CLASSIFICATION_MULTI",
    "PDF_OBJECT_DETECTION_RECTANGLE",
    "PDF_NAMED_ENTITY_RECOGNITION",
    "SPEECH_TO_TEXT",
    "TEXT_CLASSIFICATION_SINGLE",
    "TEXT_CLASSIFICATION_MULTI",
    "TEXT_TRANSCRIPTION",
    "TEXT_NER",
    "TIME_SERIES",
    "VIDEO_CLASSIFICATION_SINGLE",
    "VIDEO_OBJECT_DETECTION",
    "VIDEO_FRAME_CLASSIFICATION",
    "VIDEO_FRAME_OBJECT_TRACKING",
]


Right = Literal[
    "CAN_ACCESS_SMART_TOOLS",
    "CAN_LABEL",
    "CREATE_AUDIO",
    "CREATE_FRAME",
    "CREATE_IMAGE",
    "CREATE_PDF",
    "CREATE_TEXT",
    "CREATE_TIMESERIES",
    "CREATE_VIDEO",
    "MAKE_PUBLIC_PROJECT",
    "SEE_LICENSE_BANNER",
    "SEE_UPGRADE_BUTTON",
    "UPLOAD_CLOUD_DATA",
    "USE_API",
    "USE_API_PRIORITY",
]
